import tempfile
import os
import os.path
import requests
import json
import re
import logging
import hashlib
import inspect
import urllib.parse
import io
import pandas as pd
import tempfile

from .mixin import AttributeMixin
from .exceptions import AnaliticoException
from .status import STATUS_FAILED

import analitico.utilities
from analitico.dataset import Dataset
from analitico.utilities import id_generator

# read http streams in chunks
HTTP_BUFFER_SIZE = 32 * 1024 * 1024  # 32 MiBs


class Factory(AttributeMixin):
    """ A base class providing runtime services like notebook and plugin creation, storage, network, etc """

    def __init__(self, token=None, endpoint=None, **kwargs):
        super().__init__(**kwargs)
        if token:
            assert token.startswith("tok_")
            self.set_attribute("token", token)
        if endpoint:
            assert endpoint.startswith("http")
            self.set_attribute("endpoint", endpoint)

        # use current working directory at the time when the factory
        # is created so that the caller can setup a temp directory we
        # should work in
        self._artifacts_directory = os.getcwd()

    ##
    ## Properties and factory context
    ##

    @property
    def logger(self):
        """ Returns logger wrapped into an adapter that adds contextual information from the factory """
        return self.get_logger()

    @property
    def job(self):
        """ Job running on the server (optional) """
        return self.get_attribute("job")

    @property
    def workspace(self):
        """ Workspace context in which this factory runs (optional) """
        return self.get_attribute("workspace")

    @property
    def token(self):
        """ API token used to call endpoint (optional) """
        return self.get_attribute("token")

    @property
    def endpoint(self):
        """ Endpoint used to call analitico APIs """
        return self.get_attribute("endpoint")

    @property
    def request(self):
        """ Request used as context when running on the server or running async jobs (optional) """
        return self.get_attribute("request")

    ##
    ## Temp and cache directories
    ##

    # Temporary directory which is deleted when factory is disposed
    _temp_directory = None

    # Artifacts end up in the current working directory
    _artifacts_directory = None

    def get_temporary_directory(self):
        """ Temporary directory that can be used while a factory is used and deleted afterwards """
        temp_dir = os.path.join(tempfile.gettempdir(), "analitico_temp")
        if not os.path.isdir(temp_dir):
            os.mkdir(temp_dir)
        return temp_dir

    def get_artifacts_directory(self):
        """ 
        A plugin or notebook can produce various file artifacts during execution and place
        them in this directory (datasets, statistics, models, etc). A subclass, for example
        a factory used to run pipelines on the server, may persist files created here to cloud, etc.
        """
        return self._artifacts_directory

    def get_cache_directory(self):
        """ Returns directory to be used for caches """
        cache_dir = os.path.join(tempfile.gettempdir(), "analitico_cache")
        if not os.path.isdir(cache_dir):
            os.mkdir(cache_dir)
        return cache_dir

    def get_cache_filename(self, unique_id):
        """ Returns the fullpath in cache for an item with the given unique_id (eg: a unique url, an md5 or etag, etc) """
        # Tip: if cache contents need to be invalidated for whatever reason, you can change the prefix below...
        return os.path.join(self.get_cache_directory(), "cache_v2_" + hashlib.sha256(unique_id.encode()).hexdigest())

    ##
    ## URL retrieval, authorization and caching
    ##

    # regular expression used to detect assets using analitico:// scheme
    ANALITICO_ASSET_RE = r"(analitico://workspaces/(?P<workspace_id>[-\w.]{4,256})/)"

    # TODO could check if it would be possible to switch to an external caching library, eg:
    # https://github.com/ionrock/cachecontrol

    def get_cached_stream(self, stream, unique_id):
        """ Will cache a stream on disk based on a unique_id (like md5 or etag) and return file stream and filename """
        cache_file = self.get_cache_filename(unique_id)
        if not os.path.isfile(cache_file):
            # if not cached already, download and cache
            cache_temp_file = cache_file + ".tmp_" + id_generator()

            with open(cache_temp_file, "wb") as f:
                if hasattr(stream, "read"):
                    for chunk in iter(lambda: stream.read(HTTP_BUFFER_SIZE), b""):
                        f.write(chunk)
                else:
                    for b in stream:
                        f.write(b)

            # TODO add progress bar for slow downloads https://github.com/tqdm/tqdm#iterable-based

            os.rename(cache_temp_file, cache_file)
        # return stream from cached file
        return open(cache_file, "rb"), cache_file

    def get_url_stream(self, url, binary=False, cache=True):
        """
        Returns a stream to the given url. This works for regular http:// or https://
        and also works for analitico:// assets which are converted to calls to the given
        endpoint with proper authorization tokens. The stream is returned as an iterator.
        """
        assert url and isinstance(url, str)
        # If the url uses the analitico:// scheme for assets stored on the cloud
        # service, it will convert the url to a regular https:// scheme.
        # If the url points to an analitico API call, the request will have the
        # ?token= authorization token header added to it.
        # temporarily while all internal urls are updated to analitico://
        if url.startswith("workspaces/ws_"):
            url = analitico.ANALITICO_URL_PREFIX + url
        # see if assets uses analitico://workspaces/... scheme
        if url.startswith("analitico://"):
            if not self.endpoint:
                raise Exception("Factory is not configured with a valid API endpoint and cannot get: " + url)
            url = self.endpoint + url[len("analitico://") :]

        try:
            url_parse = urllib.parse.urlparse(url)
        except Exception:
            pass
        if url_parse and url_parse.scheme in ("http", "https"):
            headers = {}
            if url_parse.hostname and url_parse.hostname.endswith("analitico.ai") and self.token:
                # if url is connecting to analitico.ai add token
                headers = {"Authorization": "Bearer " + self.token}

            # we should not take the raw response stream here as it could be gzipped or encoded.
            # we take the decoded content as a text string and turn it into a stream or we take the
            # decompressed binary content and also turn it into a stream.
            response = requests.get(url, stream=True, headers=headers)

            # always treat content as binary, utf-8 encoding is done by readers
            response_stream = io.BytesIO(response.content)

            if cache and "etag" in response.headers:
                etag = response.headers["etag"]
                if etag:
                    return self.get_cached_stream(response_stream, url + etag)[0]
            return response_stream
        return open(url, "rb")

    def get_url_json(self, url):
        assert url and isinstance(url, str)
        url_stream = self.get_url_stream(url)
        return json.load(url_stream, encoding="utf-8")

    ##
    ## Plugins
    ##

    # dictionary of registered plugins name:class
    __plugins = {}

    @staticmethod
    def register_plugin(plugin):
        if inspect.isabstract(plugin):
            print("Factory.register_plugin: %s is abstract and cannot be registered" % plugin.Meta.name)
            return
        if plugin.Meta.name not in Factory.__plugins:
            Factory.__plugins[plugin.Meta.name] = plugin
            # print("Plugin: %s registered" % plugin.Meta.name)

    def get_plugin(self, name: str, **kwargs):
        """
        Create a plugin given its name and the environment it will run in.
        Any additional parameters passed to this method will be passed to the
        plugin initialization code and will be stored as a plugin setting.
        """
        try:
            # deprecated, temporary retrocompatibility 2019-02-24
            if name == "analitico.plugin.AugmentDatesDataframePlugin":
                name = "analitico.plugin.AugmentDatesPlugin"
            if name not in Factory.__plugins:
                self.exception("Factory.get_plugin - %s is not a registered plugin", name)
            return (Factory.__plugins[name])(factory=self, **kwargs)
        except Exception as exc:
            self.exception("Factory.get_plugin - error while creating " + name, exception=exc)

    def run_plugin(self, *args, settings, **kwargs):
        """ 
        Runs a plugin and returns its results. Takes a number of positional and named arguments
        which are passed to the plugin for execution and a dictionary of settings used to create
        the plugin. If settings are passed as an array, the method will create a pipeline plugin
        which will execute the plugins in a chain.
        """
        if isinstance(settings, list):
            settings = {"name": "analitico.plugin.PipelinePlugin", "plugins": settings}
        plugin = self.get_plugin(**settings)
        return plugin.run(*args, **kwargs)

    def get_plugins(self):
        """ Returns a list of registered plugin classes """
        return Factory.__plugins

    ##
    ## Factory methods
    ##

    EMAIL_RE = r"[^@]+@[^@]+\.[^@]+"  # very rough check

    def get_item_type(self, item_id):
        """ Returns item class from item id, eg: returns 'dataset' from ds_xxx """
        assert item_id and isinstance(item_id, str)
        if item_id.startswith(analitico.DATASET_PREFIX):
            return analitico.DATASET_TYPE
        if item_id.startswith(analitico.ENDPOINT_PREFIX):
            return analitico.ENDPOINT_TYPE
        if item_id.startswith(analitico.JOB_PREFIX):
            return analitico.JOB_TYPE
        if item_id.startswith(analitico.MODEL_PREFIX):
            return analitico.MODEL_TYPE
        if item_id.startswith(analitico.NOTEBOOK_PREFIX):
            return analitico.NOTEBOOK_TYPE
        if item_id.startswith(analitico.PLUGIN_PREFIX):
            return analitico.PLUGIN_TYPE
        if item_id.startswith(analitico.RECIPE_PREFIX):
            return analitico.RECIPE_TYPE
        if item_id.startswith(analitico.WORKER_PREFIX):
            return analitico.WORKER_TYPE
        if item_id.startswith(analitico.WORKSPACE_PREFIX):
            return analitico.WORKSPACE_TYPE
        if re.match(self.EMAIL_RE, item_id):
            return analitico.USER_TYPE
        self.warning("Factory.get_item_type - couldn't find type for: " + item_id)
        return None

    def get_item(self, item_id):
        """ Retrieves item from the server by item_id """
        assert item_id and isinstance(item_id, str) and self.endpoint.endswith("/")
        url = "{}{}s/{}".format(self.endpoint, self.get_item_type(item_id), item_id)
        return self.get_url_json(url)

    def get_dataset(self, dataset_id):
        """ Creates a Dataset object from the cloud dataset with the given id """
        plugin_settings = {
            "type": "analitico/plugin",
            "name": "analitico.plugin.CsvDataframeSourcePlugin",
            "source": {"type": "text/csv", "url": "analitico://datasets/{}/data/csv".format(dataset_id)},
        }
        # Instead of creating a plugin that reads the end product of the dataset
        # pipeline we should consider reading the dataset information from its endpoint,
        # getting the entire plugin chain and recreating it here exactly the same so it
        # can be run in Jupyter with all its plugins, etc.
        plugin = self.get_plugin(**plugin_settings)
        return Dataset(self, plugin=plugin)

    ##
    ## Logging
    ##

    class LogAdapter(logging.LoggerAdapter):
        """ A simple adapter which will call "process" on every log record to enrich it with contextual information from the Factory. """

        factory = None

        def __init__(self, logger, factory):
            super().__init__(logger, {})
            self.factory = factory

        def process(self, msg, kwargs):
            return self.factory.process_log(msg, kwargs)

    def process_log(self, msg, kwargs):
        """ Moves any kwargs other than 'exc_info' and 'extra' to 'extra' dictionary. """
        if "extra" not in kwargs:
            kwargs["extra"] = {}

        extra = kwargs["extra"]
        for key in kwargs.copy().keys():
            if key not in ("exc_info", "extra"):
                extra[key] = kwargs.pop(key)

        for attr_name in ("workspace", "token", "endpoint", "request", "job"):
            attr = self.get_attribute(attr_name, None)
            if attr:
                extra[attr_name] = attr
        return msg, kwargs

    def set_logger_level(self, level):
        """ Sets logger level to given level for all future log calls make through the factory logger """
        self.set_attribute("logger_level", level)

    def get_logger(self, name="analitico"):
        """ Returns logger wrapped into an adapter that adds contextual information from the Factory """
        logger_level = self.get_attribute("logger_level", logging.INFO)
        logger = Factory.LogAdapter(logging.getLogger(name), self)
        logger.setLevel(logger_level)
        return logger

    def debug(self, msg, *args, **kwargs):
        self.logger.log(logging.DEBUG, msg, *args, **kwargs)

    def info(self, msg, *args, **kwargs):
        self.logger.log(logging.INFO, msg, *args, **kwargs)

    def warning(self, msg, *args, **kwargs):
        self.logger.log(logging.WARNING, msg, *args, **kwargs)

    def error(self, msg, *args, **kwargs):
        self.logger.log(logging.ERROR, msg, *args, **kwargs)

    def status(self, item, status, **kwargs):
        """ Updates on the status of an item. Status is one of: created, running, canceled, completed or failed. """
        level = logging.ERROR if status == STATUS_FAILED else logging.INFO
        self.logger.log(level, "status: %s, name: %s", status, type(item).__name__, item=item, status=status, **kwargs)

    def exception(self, msg, *args, **kwargs):
        message = msg % (args)
        self.error(msg, *args, **kwargs)
        exception = kwargs.get("exception", None)
        if exception:
            raise AnaliticoException(message, **kwargs) from exception
        raise AnaliticoException(message, **kwargs)

    ##
    ## with Factory as: lifecycle methods
    ##

    def __enter__(self):
        return self

    def __exit__(self, exception_type, exception_value, traceback):
        """ Leave any temporary files upon exiting """
        pass

    ##
    ## SDK utility methods
    ##

    def get_dataframe(self, item_id: str) -> pd.DataFrame:
        """ 
        Returns the processed data from the given dataset in Analitico as a pandas DataFrame object. 
        """
        df = self.run_plugin(settings={"name": "analitico.plugin.DatasetSourcePlugin", "dataset_id": item_id})
        return df

    def upload(self, item_id: str, df: pd.DataFrame, filename: str = "data.parquet"):

        # if not (filename.startswith("data/") or filename.startswith("assets/")):
        #    raise AnaliticoException("Filename should start with assets/ or data/")

        item_type = self.get_item_type(item_id)
        url = f"{self.endpoint}{item_type}s/{item_id}/assets/{filename}"

        if filename.endswith(".parquet"):
            with tempfile.TemporaryDirectory() as tmpdir:
                filepath = os.path.join(tmpdir, filename)
                df.to_parquet(filepath)
                response = requests.post(
                    url, files={"file": filepath}, headers={"Authorization": "Bearer " + self.token}
                )
                if response.status_code != 201:
                    raise AnaliticoException(
                        f"Asset {filename} could not be uploaded to item {item_id}", response=response.json()
                    )
        else:
            raise AnaliticoException(f"Did not recognize file format for {filename}")
        return self.get_item(item_id)



import pandas as pd
import analitico.pandas
import analitico.schema

from .interfaces import IDataframePlugin, plugin

##
## AugmentDatesPlugin - dataframe in, dataframe out with better dates
##

ERROR_WHILE_AUGMENTING = "An error occoured while augmenting columns"
ERROR_WHILE_AUGMENTING_COL = "An error occoured while augmenting column: %s"
ERROR_COL_NOT_FOUND = "AugmentDatesPlugin - column '%s' was not found."


@plugin
class AugmentDatesPlugin(IDataframePlugin):
    """ A plugin that takes dates and turns them into .year, .month, .day, .dayofweek, .hour and .min columns. """

    class Meta(IDataframePlugin.Meta):
        name = "analitico.plugin.AugmentDatesPlugin"
        title = "AugmentDatesPlugin"
        description = "A plugin used to expand datetime columns into year, month, day, dayofweek, hour and minutes."
        configurations = [
            {
                "name": "schema",
                "type": "analitico/schema",
                "optional": True,
                "description": "A schema can be passed to indicate which columns should be augmented. If no schema is passed, the plugin will augment all datatime columns in the dataframe.",
            }
        ]

    def run(self, *args, action=None, **kwargs):
        try:
            df = args[0]
            if df is not None and isinstance(df, pd.DataFrame):
                columns = self.get_attribute("schema.columns")
                if columns:
                    # if columns were specified act only on those columns
                    for column in columns:
                        if "name" in column:
                            try:
                                column_name = column["name"]
                                if column["name"] in df:
                                    analitico.pandas.pd_cast_datetime(df, column_name)
                                    analitico.pandas.pd_augment_date(df, column_name)
                            except Exception as exc:
                                self.exception(ERROR_WHILE_AUGMENTING_COL, column_name, exception=exc)

                        else:
                            self.warning(ERROR_COL_NOT_FOUND, column_name)
                else:
                    # if schema was not specified just scan all columns and expand those that are datetime
                    for column in df.columns:
                        if df[column].dtype.name == analitico.schema.PD_TYPE_DATETIME:
                            analitico.pandas.pd_augment_date(df, column)
            return df
        except Exception as exc:
            self.exception(ERROR_WHILE_AUGMENTING, exception=exc)



import sys

from analitico import logger


class AnaliticoException(Exception):
    """ Base exception used in the project that can carry extra information with it in the form of a dictionary """

    default_message = "An error occurred."
    default_code = "error"
    default_status_code = 500

    message = None
    code = None
    status_code = None
    extra = None

    def __init__(self, message, *args, code=None, status_code=None, extra=None, **kwargs):
        try:
            self.message = message % (args)
        except:
            self.message = message if message else self.default_message

        self.code = code if code else self.default_code

        # we should pass the http status code by using status_code
        # however there are other classes that use status so this is
        # easy to confuse. use whatever parameter was passed.
        if status_code:
            self.status_code = status_code
        elif kwargs.get("status"):
            self.status_code = kwargs["status"]
        else:
            self.status_code = self.default_status_code

        self.extra = extra if extra else {}
        for key, value in kwargs.items():
            self.extra[key] = value

        logger.error(self)

    def __str__(self):
        return f"AnaliticoException - message: {self.message}, status_code: {self.status_code}"



import unittest
import tempfile
import numpy as np
import os


from sklearn import datasets
from sklearn import linear_model
import sklearn
import statistics

import pandas

from analitico.metadata import *


class MetadataTests(unittest.TestCase):
    """ Test utilities to write metrics and scores to training.json """

    def setUp(self):
        if os.path.exists(METADATA_FILENAME):
            os.remove(METADATA_FILENAME)

    def tearDown(self):
        if os.path.exists(METADATA_FILENAME):
            os.remove(METADATA_FILENAME)

    def test_metadata_set_score(self):
        set_metric("number_of_lines", 100)

        metadata = get_metadata()
        self.assertEqual(len(metadata["scores"]), 1)
        self.assertEqual(metadata["scores"]["number_of_lines"], 100)

    def test_metadata_set_score_with_title(self):
        set_metric("number_of_lines2", 100, title="Number of lines", subtitle="A longer description")

        metadata = get_metadata()
        self.assertEqual(len(metadata["scores"]), 1)
        self.assertEqual(metadata["scores"]["number_of_lines2"]["value"], 100)
        self.assertEqual(metadata["scores"]["number_of_lines2"]["title"], "Number of lines")
        self.assertEqual(metadata["scores"]["number_of_lines2"]["subtitle"], "A longer description")

    def test_metadata_set_score_with_title_and_priority(self):
        set_metric("number_of_lines3", 100, title="Number of lines", priority=1)

        metadata = get_metadata()
        self.assertEqual(len(metadata["scores"]), 1)
        self.assertEqual(len(metadata["scores"]["number_of_lines3"]), 3)
        self.assertEqual(metadata["scores"]["number_of_lines3"]["value"], 100)
        self.assertEqual(metadata["scores"]["number_of_lines3"]["title"], "Number of lines")
        self.assertEqual(metadata["scores"]["number_of_lines3"]["priority"], 1)

    def test_metadata_set_multiple_scores(self):
        set_metric("metric1", 100)
        set_metric("metric2", "hello")

        metadata = get_metadata()
        self.assertEqual(metadata["scores"]["metric1"], 100)
        self.assertEqual(metadata["scores"]["metric2"], "hello")

    def test_metadata_set_category_scores_with_category(self):
        set_metric("metric1", 100)
        set_metric("metric2", "hello", category="category2")

        metadata = get_metadata()
        self.assertEqual(len(metadata["scores"]), 2)
        self.assertEqual(metadata["scores"]["metric1"], 100)
        self.assertEqual(metadata["scores"]["category2"]["metric2"], "hello")

    def test_metadata_set_category_scores_with_category_title(self):
        set_metric("metric1", 100)
        set_metric("metric2", "hello", category="category2", category_title="Category Two")

        metadata = get_metadata()
        self.assertEqual(len(metadata["scores"]), 2)
        self.assertEqual(metadata["scores"]["metric1"], 100)
        self.assertEqual(metadata["scores"]["category2"]["title"], "Category Two")
        self.assertEqual(metadata["scores"]["category2"]["metric2"], "hello")

    def test_metadata_set_category_scores_with_category_title_and_scores_extras(self):
        set_metric("metric1", 100)
        set_metric(
            "metric2",
            "hello",
            priority=1,
            title="VIP Metric",
            category="category2",
            category_title="Category Two",
            category_subtitle="A longer subtitle",
        )

        metadata = get_metadata()
        self.assertEqual(len(metadata["scores"]), 2)
        self.assertEqual(metadata["scores"]["metric1"], 100)

        self.assertEqual(metadata["scores"]["category2"]["title"], "Category Two")
        self.assertEqual(metadata["scores"]["category2"]["subtitle"], "A longer subtitle")
        self.assertEqual(metadata["scores"]["category2"]["metric2"]["value"], "hello")
        self.assertEqual(metadata["scores"]["category2"]["metric2"]["priority"], 1)
        self.assertEqual(metadata["scores"]["category2"]["metric2"]["title"], "VIP Metric")

    def test_metadata_scores_scikit_regressor(self):
        data = sklearn.datasets.load_boston()
        df = pandas.DataFrame(data.data, columns=data.feature_names)
        target = pandas.DataFrame(data.target, columns=["MEDV"])

        X = df
        y_true = target["MEDV"]
        lm = linear_model.LinearRegression()
        model = lm.fit(X, y_true)
        y_pred = lm.predict(X)

        # save a variety of scores derived from model, data and predictions
        set_model_metrics(y_true, y_pred, model=model)

        # check regressor scores were saved correctly
        scores = get_metadata()["scores"]["sklearn_metrics"]
        self.assertIn("mean_abs_error", scores)
        self.assertAlmostEqual(scores["mean_abs_error"]["value"], 3.27086, 2)
        self.assertIn("mean_squared_error", scores)
        self.assertAlmostEqual(scores["mean_squared_error"]["value"], 21.89483, 2)
        self.assertIn("median_abs_error", scores)
        self.assertAlmostEqual(scores["median_abs_error"]["value"], 2.45231, 2)

    def test_metadata_scores_scikit_classificator(self):
        """ Test extracting metrics metadata from a scikit learn classificator. """
        iris = sklearn.datasets.load_iris()
        classificator = sklearn.svm.SVC(gamma="scale")

        X, y_true = iris.data, iris.target
        model = classificator.fit(X, y_true)
        y_pred = model.predict(X)

        # save a variety of scores derived from model, data and predictions
        set_model_metrics(y_true, y_pred, target_names=iris.target_names, model=model)

        # check regressor scores were saved correctly
        scores = get_metadata()["scores"]["sklearn_metrics"]

        self.assertIn("classification_report", scores)
        for target_name in iris.target_names:
            self.assertIn(target_name, scores["classification_report"]["value"])

        self.assertIn("confusion_matrix", scores)

        self.assertAlmostEqual(scores["accuracy_score"]["value"], 0.98667, 2)
        self.assertAlmostEqual(scores["precision_score_macro"]["value"], 0.98718, 2)
        self.assertAlmostEqual(scores["precision_score_micro"]["value"], 0.98667, 2)
        self.assertAlmostEqual(scores["precision_score_weighted"]["value"], 0.98718, 2)

        # not a binary classifier, some metrics should not be there
        self.assertNotIn("log_loss", scores)



import pandas as pd

from analitico.pandas import pd_columns_to_string
from .interfaces import PluginError, IDataframePlugin
from .pipelineplugin import PipelinePlugin, plugin

##
## FusionDataframePlugin
##

ERROR_NO_INPUT_DF = "Should receive as input a single pd.DataFrame, received: %s"
ERROR_NO_PIPELINE_DF = (
    "Plugins pipeline should produce a pd.DataFrame to be merged with main input, instead received: %s"
)
ERROR_NO_LEFT_COLUMN = (
    "The left (main) dataframe does not have a column named '%s' to merge on. Available columns are: %s."
)
ERROR_NO_RIGHT_COLUMN = (
    "The right (secondary) dataframe does not have a column named '%s' to merge on. Available columns are: %s."
)
ERROR_NO_MERGE_CONF = "You need to specify how to merge dataframes either with the 'on' attribute or with the 'left_on' and 'right_on' attributes indicating the column names"


@plugin
class FusionDataframePlugin(PipelinePlugin):
    """ 
    A plugin used to merge two datasources using specific merge rules.
    This plugin is a IDataframePlugin because it takes a dataframe
    (the main table or left table), modifies it (join) and returns it.
    It is also a DataframePipelinePlugin because it can embed a second
    pipeline which generates the secondary, or right, table which is
    merged with the main. Merging is performed based on rules described 
    in the "merge" attribute, which is a dictionary that closely maps
    pandas' merge parameters.
    """

    class Meta(IDataframePlugin.Meta):
        name = "analitico.plugin.FusionDataframePlugin"
        inputs = [{"name": "dataframe", "type": "pandas.DataFrame"}]
        outputs = [{"name": "dataframe", "type": "pandas.DataFrame"}]

    def run(self, *args, action=None, **kwargs) -> pd.DataFrame:
        """ Merge two pipelines into a single dataframe """
        try:
            df_left = args[0]
            if not isinstance(df_left, pd.DataFrame):
                self.exception(ERROR_NO_INPUT_DF, df_left)

            # run the pipeline to obtain the secondary table (right) that we're joining on
            df_right = super().run(action=action, **kwargs)
            if not isinstance(df_right, pd.DataFrame):
                self.exception(ERROR_NO_PIPELINE_DF, df_right)

            self.info("Left columns: %s", pd_columns_to_string(df_left))
            self.info("Left rows: %d", len(df_left))
            self.info("Right columns: %s", pd_columns_to_string(df_right))
            self.info("Right rows: %d", len(df_right))

            # "merge" attribute contains a dictionary of settings that closely match those of pandas.merge:
            # https://pandas.pydata.org/pandas-docs/stable/user_guide/merging.html#database-style-dataframe-or-named-series-joining-merging

            merge = self.get_attribute("merge")
            if not merge:
                self.exception("Attribute 'merge' with merging details is required")

            # "how" determines how we merge
            how = merge.get("how", "inner")
            how_options = ["left", "right", "outer", "inner"]
            if how not in how_options:
                self.exception("Attribute how: %s is unknown, should be one of %s", how, str(how_options))

            on = merge.get("on", None)
            if on:
                self.info("Merge on: %s", on)

                if on not in df_left.columns:
                    self.exception(ERROR_NO_LEFT_COLUMN, on, pd_columns_to_string(df_left))
                if on not in df_right.columns:
                    self.exception(ERROR_NO_RIGHT_COLUMN, on, pd_columns_to_string(df_right))

                df_fusion = pd.merge(df_left, df_right, on=on, how=how)

            else:
                left_on = merge.get("left_on", None)
                right_on = merge.get("right_on", None)

                if left_on and right_on:
                    self.info("Merge left_on: %s, right_on: %s", left_on, right_on)
                    if left_on not in df_left.columns:
                        self.exception(ERROR_NO_LEFT_COLUMN, left_on, pd_columns_to_string(df_left))
                    if right_on not in df_right.columns:
                        self.exception(ERROR_NO_RIGHT_COLUMN, right_on, pd_columns_to_string(df_right))

                    df_fusion = pd.merge(df_left, df_right, left_on=left_on, right_on=right_on, how=how)
                else:
                    self.exception(ERROR_NO_MERGE_CONF)

            return df_fusion

        except PluginError as plugin_error:
            raise plugin_error
        except Exception as exc:
            self.exception("Exception while merging dataframes", exception=exc)



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()


setuptools.setup(
    name="analitico",
    version="2019.07.14.dev",
    author="Analitico Labs Inc.",
    author_email="info@analitico.ai",
    description="A Python package for Analitico.ai",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/analitico/analitico-sdk",
    packages=setuptools.find_packages(),
    package_data={
        "analitico": ["test/assets/*.*"]
        },
    install_requires=[
        "catboost>=0.12.2", 
        "numpy>=1.14.6", 
        "pandas>=0.22.0"
        ],
    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "Environment :: Console",
        "Intended Audience :: Science/Research",
        "Intended Audience :: Developers",
        "Intended Audience :: Financial and Insurance Industry",
        "Intended Audience :: Information Technology",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
        "Topic :: Scientific/Engineering :: Information Analysis",
    ],
)



import os
import time
import pandas as pd
import logging
import socket
import platform
import multiprocessing
import psutil
import collections
import subprocess
import sys
import random
import string
import dateutil
import re
import subprocess
import traceback
import shutil

from cryptography.hazmat.primitives import serialization as crypto_serialization
from cryptography.hazmat.primitives.asymmetric import rsa
from cryptography.hazmat.backends import default_backend as crypto_default_backend

from collections import OrderedDict

# use simplejson instead of standard built in library
# mostly because it has a parameter which supports replacing nan with nulls
# thus producing json which is ecma compliant and won't have issues being read
# https://simplejson.readthedocs.io/en/latest/
import simplejson as json

from datetime import datetime

try:
    import distro
    import GPUtil
except Exception:
    pass

from analitico.exceptions import AnaliticoException
from analitico.schema import analitico_to_pandas_type, apply_schema, NA_VALUES

# default logger for analitico's libraries
logger = logging.getLogger("analitico")


# RESTful API Design Tips from Experience
# https://medium.com/studioarmix/learn-restful-api-design-ideals-c5ec915a430f

# Trying to follow this spec for naming, etc
# https://jsonapi.org/format/#document-top-level

# Following this format for errors:
# https://jsonapi.org/format/#errors


##
## Exceptions
##


def first_or_list(items):
    try:
        if items and len(items) == 1:
            return items[0]
    except Exception:
        pass  # validation errors pass a dictionary so we just pass it through
    return items


def exception_to_dict(exception: Exception, add_context=True, add_formatted=True, add_traceback=True) -> dict:
    """ Returns a dictionary with detailed information on the given exception and its inner (chained) exceptions """

    # trying to adhere as much as possible to json:api specs here
    # https://jsonapi.org/format/#errors
    d = OrderedDict()
    d["status"] = "500"  # want this to go first
    d["code"] = type(exception).__name__.lower()
    d["title"] = str(exception.args[0]) if len(exception.args) > 0 else str(exception)
    d["meta"] = {}

    if isinstance(exception, AnaliticoException):
        d["status"] = str(exception.status_code)
        d["code"] = exception.code
        d["title"] = exception.message
        if exception.extra and len(exception.extra) > 0:
            d["meta"]["extra"] = json_sanitize_dict(exception.extra)

    if add_context and exception.__context__:
        d["meta"]["context"] = exception_to_dict(
            exception.__context__, add_context=True, add_formatted=False, add_traceback=False
        )

    # information on exception currently being handled
    _, _, exc_traceback = sys.exc_info()

    if add_formatted:
        # printout of error condition
        d["meta"]["formatted"] = traceback.format_exception(type(exception), exception, exc_traceback)

    if add_traceback:
        # extract frame summaries from traceback and convert them
        # to list of dictionaries with file and line number information
        d["meta"]["traceback"] = []
        for fs in traceback.extract_tb(exc_traceback, 20):
            d["meta"]["traceback"].append(
                OrderedDict(
                    {
                        "summary": "File '{}', line {}, in {}".format(fs.filename, fs.lineno, fs.name),
                        "filename": fs.filename,
                        "line": fs.line,
                        "lineno": fs.lineno,
                        "name": fs.name,
                    }
                )
            )

    if d["status"] is None:
        d.pop("status")
    if len(d["meta"]) < 1:
        d.pop("meta")
    return d


##
## Files
##


def copy_directory(src_directory, dst_directory):
    """
    Copy source directory recursively into destination directory.
    Works like shutil.copytree but does not require that target directory does not exist.
    If a file already exists at the destination it will be overwritten.
    Symbolic links in the source directory are followed.
    
    Arguments:
        src_directory {str} -- The source directory.
        dst_directory {str} -- The destination directory.
    """
    if not src_directory.endswith("/"): src_directory += "/"
    if not dst_directory.endswith("/"): dst_directory += "/"
    if not os.path.exists(dst_directory): os.makedirs(dst_directory)
    # wthout shell=True the star char is parsed by subprocess
    subprocess_run(cmd_args=[f"cp -rfH {src_directory}* {dst_directory}"], shell=True)


##
## Crypto
##


def id_generator(size=8, chars="abcdefghijklmnopqrstuvwxyz0123456789"):
    return "".join(random.choice(chars) for _ in range(size))


def ssh_key_generator():
    """ Generate ssh private and public key """
    key = rsa.generate_private_key(backend=crypto_default_backend(), public_exponent=65537, key_size=2048)
    private_key = key.private_bytes(
        crypto_serialization.Encoding.PEM, crypto_serialization.PrivateFormat.PKCS8, crypto_serialization.NoEncryption()
    )
    public_key = key.public_key().public_bytes(
        crypto_serialization.Encoding.OpenSSH, crypto_serialization.PublicFormat.OpenSSH
    )
    return (private_key, public_key)


##
## Runtime
##

MB = 1024 * 1024


def get_runtime_brief():
    """ A digest version of get_runtime to be used more frequently """
    return {"cpu_count": multiprocessing.cpu_count()}


def get_gpu_runtime():
    """ Returns array of GPU specs (if discoverable) """
    try:
        runtime = []
        # optional package
        GPUs = GPUtil.getGPUs()
        if GPUs:
            for GPU in GPUs:
                runtime.append(
                    {
                        "uuid": GPU.uuid,
                        "name": GPU.name,
                        "driver": GPU.driver,
                        "temperature": int(GPU.temperature),
                        "load": round(GPU.load, 2),
                        "memory": {
                            "total_mb": int(GPU.memoryTotal),
                            "available_ms": int(GPU.memoryFree),
                            "used_mb": int(GPU.memoryUsed),
                            "used_perc": round(GPU.memoryUtil, 2),
                        },
                    }
                )
    except Exception:
        return None
    return runtime


def get_runtime():
    """ Collect information on runtime environment, platform, python, hardware, etc """
    started_on = time_ms()
    runtime = collections.OrderedDict()
    try:
        runtime["hostname"] = socket.gethostname()
        runtime["ip"] = socket.gethostbyname(socket.gethostname())
        runtime["platform"] = {"system": platform.system(), "version": platform.version()}
        try:
            # optional package
            runtime["platform"]["name"] = distro.name()
            runtime["platform"]["version"] = distro.version()
        except Exception:
            pass

        runtime["python"] = {
            "version": platform.python_version(),
            "implementation": platform.python_implementation(),
            "encoding": sys.getdefaultencoding(),
        }

        hardware = collections.OrderedDict()
        runtime["hardware"] = hardware
        hardware["cpu"] = {"type": platform.processor(), "count": multiprocessing.cpu_count()}
        try:
            # will raise exception on virtual machines
            hardware["cpu"]["freq"] = int(psutil.cpu_freq()[2])
        except:
            pass

        gpu = get_gpu_runtime()
        if gpu:
            hardware["gpu"] = gpu

        memory = psutil.virtual_memory()
        swap = psutil.swap_memory()
        hardware["memory"] = {
            "total_mb": int(memory.total / MB),
            "available_mb": int(memory.available / MB),
            "used_mb": int(memory.used / MB),
            "swap_mb": int(swap.total / MB),
            "swap_perc": round(swap.percent, 2),
        }

        disk = psutil.disk_usage("/")
        hardware["disk"] = {
            "total_mb": int(disk.total / MB),
            "available_mb": int(disk.free / MB),
            "used_mb": int(disk.used / MB),
            "used_perc": round(disk.percent, 2),
        }

        boot_time = datetime.fromtimestamp(psutil.boot_time())
        uptime = (datetime.now() - boot_time).total_seconds() / 3600
        runtime["uptime"] = {"since": boot_time.strftime("%Y-%m-%d %H:%M:%S"), "hours": round(uptime, 2)}

        # production servers have an environment variable indicating git commit
        runtime["github"] = {}
        try:
            v = subprocess.check_output(["git", "describe"]).decode("utf-8").strip()
            runtime["github"]["version"] = v
        except:
            pass
        commit_sha = os.environ.get("ANALITICO_COMMIT_SHA", None)
        if commit_sha:
            runtime["github"]["commit"] = commit_sha
            runtime["github"]["url"] = "https://github.com/analitico/analitico/commit/" + commit_sha
    except Exception as exc:
        runtime["exception"] = str(exc)

    runtime["elapsed_ms"] = time_ms(started_on)
    return runtime


##
## Json utilities
##

JSON_NOT_SERIALIZABLE = "NOT_SERIALIZABLE"


def json_sanitize_dict(dict):
    """ Remove from dictionary all items which cannot be easily serialized to json, replace with item_id where possible. """
    sanitized = json.loads(json.dumps(dict, skipkeys=True, default=lambda o: JSON_NOT_SERIALIZABLE))
    for key in dict:
        if sanitized[key] == JSON_NOT_SERIALIZABLE:
            try:
                # if an item could note be serialized, let's see if we can replace it with its id
                key_id = key + "_id"
                if key_id not in sanitized:
                    sanitized[key_id] = dict[key].id
                    sanitized.pop(key)
            except:
                # item doesn't have an .id
                pass
    return sanitized


def save_json(data, filename, indent=None, encoding="utf8", ignore_nan=True):
    """ Saves given data in a json file, encodes as utf-8, replace np.NaN with nulls """
    with open(filename, "w", encoding=encoding) as f:
        json.dump(data, f, indent=indent, ignore_nan=ignore_nan)


def read_json(filename, encoding="utf-8"):
    """ Reads, decodes and returns the contents of a json file """
    try:
        with open(filename, encoding=encoding) as f:
            return json.load(f)
    except Exception as exc:
        detail = "analitico.utilities.read_json: error while reading {}, exception: {}".format(filename, exc)
        logger.error(detail)
        raise Exception(detail, exc)


def read_text(filename, encoding="utf-8"):
    """ Reads a text file """
    try:
        with open(filename, encoding=encoding) as f:
            return f.read()
    except Exception as exc:
        detail = "analitico.utilities.read_text: error while reading {}, exception: {}".format(filename, exc)
        logger.error(detail)
        raise Exception(detail, exc)


def save_text(text, filename):
    with open(filename, "w") as text_file:
        text_file.write(text)


##
## Time utilities
##


def time_ms(started_on=None):
    """ Returns the time elapsed since given time in ms """
    return datetime.now() if started_on is None else int((datetime.now() - started_on).total_seconds() * 1000)


# https://docs.python.org/2/library/functools.html
# used as @timeit decorator
# https://medium.com/pythonhive/python-decorator-to-measure-the-execution-time-of-methods-fa04cb6bb36d
def timeit(method):
    def timed(*args, **kwargs):
        ts = time.time()
        result = method(*args, **kwargs)
        ms = int((time.time() - ts) * 1000)

        # if self has logger, log to it
        if hasattr(args[0], "logger"):
            try:
                args[0].logger.info("%s in %d ms", method.__name__.lower(), ms)
            except:
                pass
        else:
            print("%s in %d ms", method.__name__.lower(), ms)

        return result

    return timed


##
## Timestamp utilities
##


def timestamp_to_time(ts: str, ts_format="%Y-%m-%d %H:%M:%S") -> time.struct_time:
    """ Converts a timestamp string in the given format to a time object """
    try:
        return time.strptime(ts, ts_format)
    except TypeError:
        return None


def timestamp_to_secs(ts: str, ts_format="%Y-%m-%d %H:%M:%S") -> float:
    """ Converts a timestamp string to number of seconds since epoch """
    return time.mktime(time.strptime(ts, ts_format))


def timestamp_diff_secs(ts1, ts2):
    t1 = timestamp_to_secs(ts1)
    t2 = timestamp_to_secs(ts2)
    return t1 - t2


##
## Dictionary utilities
##


def get_dict_dot(d: dict, key: str, default=None):
    """ Gets an entry from a dictionary using dot notation key, eg: this.that.something """
    try:
        if isinstance(d, dict) and key:
            split = key.split(".")
            value = d.get(split[0])
            if value:
                if len(split) == 1:
                    return value
                return get_dict_dot(value, key[len(split[0]) + 1 :], default)
    except KeyError:
        pass
    return default


def set_dict_dot(d: dict, key: str, value=None):
    """ Sets an entry from a dictionary using dot notation key, eg: this.that.something """
    if isinstance(d, dict) and key:
        split = key.split(".")
        subkey = split[0]
        if len(split) == 1:
            d[subkey] = value
            return
        if not (subkey in d):
            d[subkey] = OrderedDict()
        set_dict_dot(d[subkey], key[len(subkey) + 1 :], value)


##
## CSV
##


def get_csv_row_count(filename):
    """ Returns the number of rows in the given csv file (one row is deducted for the header) """
    with open(filename, "r") as f:
        return sum(1 for row in f) - 1


##
## Regular expression
##


def re_match_group(expression: str, content: str, default: str = None, group_index: int = 1) -> str:
    """ Searches for the first matching group or returns the default value """
    match = re.search(expression, content)
    if match:
        return match.group(group_index)
    return default


##
## Text
##


def comma_separated_to_array(items: str) -> [str]:
    """ Turns a string with a comma separated list of items into an array of strings """
    if items and items.strip():
        return [x.strip() for x in items.split(",")]
    return None


def array_to_comma_separated(items: [str]) -> str:
    """ Turns an array of items into a comma separated string """
    if items and len(items) > 0:
        return ",".join(items)
    return None


##
## Subprocess
##


def json_from_string_if_possible(value: str):
    """ If your string is json then it will be parsed and returned otherwise you just get your string back """
    try:
        return json.loads(value) if len(value) > 4 else value
    except:
        pass
    return value


def subprocess_run(cmd_args, job=None, timeout=3600, cwd=None, shell=False) -> (str, str):
    """
    Run a subprocess with the given command arguments. Logs the command, the response
    and the time it took to run it. If an error occours, raises an explanatory exception
    otherwise it returns the stdout and stderr from the command possibly parse as json
    if the response was in json.
    """
    message = "subprocess_run:\n" + " ".join(cmd_args)
    logger.info(message)
    if job:
        job.append_logs(message)

    started_on = time_ms()
    response = subprocess.run(
        cmd_args, encoding="utf-8", stdout=subprocess.PIPE, stderr=subprocess.PIPE, timeout=timeout, cwd=cwd, shell=shell
    )

    elapsed_ms = time_ms(started_on)
    message = (
        f"completed in {elapsed_ms} ms, returned code: {response.returncode}\n\n{response.stdout}\n\n{response.stderr}"
    )
    logger.info(message)
    if job:
        job.append_logs(message)

    if response.returncode:
        message = "An error occoured while executing '" + " ".join(cmd_args) + "'."
        if response.stdout:
            message += "\nResponse.stdout:\n" + response.stdout
        if response.stderr:
            message += "\nResponse.stderr:\n" + response.stderr
        logger.error(message)
        raise AnaliticoException(message)

    stdout = json_from_string_if_possible(response.stdout)
    stderr = json_from_string_if_possible(response.stderr)
    return stdout, stderr



import collections

from analitico.utilities import get_dict_dot, set_dict_dot


class AttributeMixin:
    """
    A simple mixin to implement a class with configurable attributes

    When this class or its subclass is initialized it can take any number of named
    arguments in its constructor, eg: obj = AttributesMixing(attribute1='value1', attribute2='value2')
    You can then access these attributes using obj.attribute1 or by calling obj.get_attribute
    with the name of the attribute. The purpose of the mixin is to allow for simple storage,
    retrieval and persistence of attributes in classes without having to know a priori their contents.
    This comes useful in the case of plugins and Django models for example.
    """

    # The attribute field will contain a dictionary. It can be instantiated on demand and can also be
    # implemented as a JSONField in a Django concrete model class to work around a django issue:
    # attributes = jsonfield.JSONField(load_kwargs={'object_pairs_hook': collections.OrderedDict}, blank=True, null=True)
    attributes = None

    def __init__(self, **kwargs):
        self.attributes = kwargs

    def __getattr__(self, key):
        """ Returns the setting if found or AttributeError if not found """
        if self.attributes and key in self.attributes:
            return self.get_attribute(key)
        raise AttributeError(str(self) + " has no attribute: " + key)

    def get_attribute(self, key, default=None):
        """ 
        Returns a setting if configured or the given default value if not.
        You can find the value of hierarchical settings using a dot structure
        like this.that.setting which will navigate the dictionary of settings
        to the correct level.
        """
        if self.attributes:
            return get_dict_dot(self.attributes, key, default)
        return default

    def set_attribute(self, key, value):
        """ Set the value of the given attribute """
        if not self.attributes:
            self.attributes = collections.OrderedDict()
        attributes = self.attributes
        set_dict_dot(attributes, key, value)
        # self.attributes needs to be assigned basically to itself
        # in case the superclass is a Django model which is monitoring
        # the field to decide when a remove SQL instance need to be save()d
        self.attributes = attributes



import pandas as pd
import json
import dateutil
from io import StringIO

import analitico
import analitico.schema
import analitico.utilities

from analitico import AnaliticoException, logger
from analitico.schema import analitico_to_pandas_type, NA_VALUES

##
## Pandas utilities
##


def pd_print_nulls(df):
    for column in df:
        nulls = df[column].isnull().sum()
        if nulls > 0:
            perc = 100.0 * nulls / len(df)
            print("{0} has {1} null values {2:1.2f}%".format(column, nulls, perc))


def pd_date_parser(x):
    if not x:
        return None
    lower_x = x.lower()
    if lower_x in ("none", "null", "nan", "empty"):
        return None
    date = dateutil.parser.parser(x)
    return date


def pd_cast_datetime(df: pd.DataFrame, column: str) -> pd.DataFrame:
    """ Casts a string column to a date column, assumes format is recognizable """
    df[column] = pd.to_datetime(df[column], infer_datetime_format=True, errors="coerce")
    return df


def move_column(df: pd.DataFrame, column: str, index=0) -> pd.DataFrame:
    """ Moves the column with the given name to the given index, returns dataframe """
    columns = list(df.columns)
    if column not in columns:
        logger.warning(f"move_column - column {column} is not present in df.columns: {df.columns}")
        return df
    columns.insert(index, columns.pop(columns.index(column)))
    return df[columns]


EXPAND_ALL_COLUMNS = ["dayofweek", "year", "month", "day", "hour", "minute"]


def augment_dates(df: pd.DataFrame, column: str = None, expand=None, drop=True) -> pd.DataFrame:
    """
    Augment the specific column contaning dates into a number of separate columns with the day of the week,
    year, month, day, hour and minute. If a column name is not specified, the method will expand all columns
    of type datetime. If a column is specified but it's not of type datetime, the column will be converted to
    datetime. The expanded column is the dropped from the dataframe unless otherwise specified.
    """
    try:
        if not isinstance(df, pd.DataFrame):
            raise AnaliticoException("augment_dates - requires a pd.DataFrame")

        # if a specific column was not specified we scan all columns and
        # apply date augmentation to all columns of type datetime
        if not column:
            # TODO figure out why analitico.schema.PD_TYPE_DATETIME doesn't work
            for col1 in df.columns.copy():
                if df[col1].dtype == "datetime64[ns]":
                    df = augment_dates(df, col1, expand, drop)
            return df

        if column not in df.columns:
            raise AnaliticoException(f"augment_dates - cannot find column {column} in df.columns: {df.columns}")

        if df[column].dtype != analitico.schema.PD_TYPE_DATETIME:
            logger.info(
                f"augment_dates - changing column {column} from type {df[column].dtype} to {analitico.schema.PD_TYPE_DATETIME}"
            )
            df = pd_cast_datetime(df, column)

        # TODO warn of missing date fields, log number of missing records

        if not expand:
            expand = EXPAND_ALL_COLUMNS

        loc = df.columns.get_loc(column) + 1
        dates = pd.DatetimeIndex(df[column])
        if "dayofweek" in expand:
            df[column + ".dayofweek"] = dates.dayofweek.astype("category")
            df = move_column(df, column + ".dayofweek", loc)
            loc += 1
        if "year" in expand:
            df[column + ".year"] = dates.year.astype("category")
            df = move_column(df, column + ".year", loc)
            loc += 1
        if "month" in expand:
            df[column + ".month"] = dates.month.astype("category")
            df = move_column(df, column + ".month", loc)
            loc += 1
        if "day" in expand:
            df[column + ".day"] = dates.day.astype("category")
            df = move_column(df, column + ".day", loc)
            loc += 1
        if "hour" in expand:
            df[column + ".hour"] = dates.hour.astype("category")
            df = move_column(df, column + ".hour", loc)
            loc += 1
        if "minute" in expand:
            df[column + ".minute"] = dates.minute.astype("category", copy=False)
            df = move_column(df, column + ".minute", loc)
            loc += 1

        if drop:
            df = df.drop([column], axis=1, inplace=False)

    except AnaliticoException:
        raise
    except Exception as exc:
        raise AnaliticoException(
            f"augment_dates - an error occoured while augmenting dates in column {column}"
        ) from exc

    return df


# DEPRECATED
def pd_augment_date(df, column):
    """ Augments a datetime column into year, month, day, dayofweek, hour, minute """
    if column not in df.columns:
        raise Exception("pd_augment_date - column '" + column + "' is missing")
    # create separate columns for each parameter (overwrite if needed)
    dates = pd.DatetimeIndex(df[column])
    df[column + ".year"] = dates.year.astype("category", copy=False)
    df[column + ".month"] = dates.month.astype("category", copy=False)
    df[column + ".day"] = dates.day.astype("category", copy=False)
    df[column + ".hour"] = dates.hour.astype("category", copy=False)
    df[column + ".minute"] = dates.minute.astype("category", copy=False)
    df[column + ".dayofweek"] = dates.dayofweek.astype("category", copy=False)
    # TODO place augmented columns next to original
    # loc = df.columns.get_loc(column) + 1
    # df.drop([column], axis=1, inplace=True)


def pd_timediff_min(df, column_start, column_end, column_diff):
    """ Creates a new column with difference in minutes between the two named columns """
    pd_cast_datetime(df, column_start)
    pd_cast_datetime(df, column_end)
    df[column_diff] = df[column_end] - df[column_start]
    df[column_diff] = df[column_diff].dt.total_seconds() / 60.0


def pd_columns_to_string(df):
    """ Returns a single string with a list of columns, eg: 'col1', 'col2', 'col3' """
    columns = "".join("'" + column + "', " for column in df.columns)
    return columns[:-2]


def pd_read_csv(filepath_or_buffer, schema=None):
    """ Read csv file from file or stream and apply optional schema """
    try:
        dtype = None

        if schema:
            # array of types for each column in the source
            columns = schema.get("columns")
            if columns:
                dtype = {}
                for column in columns:
                    if "type" in column:  # type is optionally defined
                        if column["type"] == "datetime":
                            dtype[column["name"]] = "object"
                        elif column["type"] == "timespan":
                            dtype[column["name"]] = "object"
                        elif column["type"] == "integer":
                            pass  # do not cast so we can deal with nulls later
                        else:
                            dtype[column["name"]] = analitico_to_pandas_type(column["type"])

        # read csv from file or stream
        df = pd.read_csv(filepath_or_buffer, dtype=dtype, encoding="utf-8", na_values=NA_VALUES, low_memory=False)

        if schema:
            # reorder, filter, apply types, rename columns as requested in schema
            df = analitico.schema.apply_schema(df, schema)
        return df

    except Exception as exc:
        logger.error(f"Could not read csv file from {filepath_or_buffer}, schema: {schema}, dtype: {dtype}")
        raise exc


def pd_to_csv(df: pd.DataFrame, filename, schema=False, samples=0):
    """ Writes dataframe to disk optionally adding a .schema file and a .samples file """
    if not filename.endswith(".csv"):
        raise Exception("pd_to_csv - filename " + filename + " should end in .csv")
    df.to_csv(filename, encoding="utf-8")
    if schema:
        schema = analitico.schema.generate_schema(df)
        schemaname = filename + ".info"
        analitico.utilities.save_json({"schema": schema}, schemaname)
    if samples > 0 and not df.empty:
        samples = pd_sample(df, samples)
        samplesname = filename[:-4] + ".samples.csv"
        samples.to_csv(samplesname, encoding="utf-8")


def pd_drop_column(df, column, inplace=False):
    """ Drops a column, no exceptions if it's not there """
    try:
        return df.drop([column], axis=1, inplace=inplace)
    except:
        logger.warning(f"pd_drop_column - could not find column {column}")
    return df


def pd_to_dict(df):
    """ Convert a dataframe to json, encodes all dates and timestamps to ISO8601 """
    assert isinstance(df, pd.DataFrame), "pd_to_dict - requires a pd.DataFrame"
    with StringIO() as io:
        df.to_json(io, orient="records", date_format="iso", date_unit="s", double_precision=6)
        io.seek(0)
        return json.load(io)


def pd_sample(df, n=20):
    """ Returns a sample from the given DataFrame, either number of rows or percentage. """
    if n < 1:
        return df.sample(frac=n)
    if n < len(df.index):
        return df.sample(n=n)
    return df



import os
import os.path
import pandas as pd

from .interfaces import PluginError
from .pipelineplugin import PipelinePlugin, plugin

import analitico.constants
import analitico.pandas

from analitico import status, AnaliticoException
from analitico.utilities import read_json

##
## RecipePipelinePlugin
##


@plugin
class RecipePipelinePlugin(PipelinePlugin):
    """
    A recipe pipeline contains:
    - a dataframe source plugin that gathers the training data
    - a miminal (or empty) set of plugins to do some final filtering (eg. remove columns)
    - an algorithm used for training (eg. classifier, neural net, etc)
    Running the pipeline produces a trained model which can be used later for predictions.
    """

    class Meta(PipelinePlugin.Meta):
        name = "analitico.plugin.RecipePipelinePlugin"
        inputs = None
        outputs = [{"model": "dict"}]

    def run(self, *args, action=None, **kwargs):
        """ Process the plugins in sequence to create trained model artifacts """
        artifacts_path = self.factory.get_artifacts_directory()
        training_path = os.path.join(artifacts_path, "training.json")

        # when training run the recipe which will produce the training artifacts
        if analitico.constants.ACTION_TRAIN in action:
            results = super().run(*args, action=action, **kwargs)
            # training.json, trained models and other artifacts should
            # now be in the artifacts directory. depending on the environment
            # these may be left on disk (SDK) or stored in cloud (APIs)
            if not isinstance(results, dict):
                self.factory.exception(
                    "Pipeline didn't return a dictionary with training results",
                    item=self,
                    artifacts_path=artifacts_path,
                )
            if not os.path.isfile(training_path):
                self.factory.exception(
                    "Pipeline didn't produce training.json", item=self, artifacts_path=artifacts_path
                )
            return results

        # when predicting pass the data to the recipe for prediction
        if analitico.constants.ACTION_PREDICT in action:
            # check for training information on disk
            if not os.path.isfile(training_path):
                self.factory.exception(
                    "Pipeline can't fine file training.json", item=self, artifacts_path=artifacts_path
                )

            # normally prediction input is a pd.DataFrame
            if isinstance(args[0], pd.DataFrame):
                df_original = args[0]  # original data
                df = df_original.copy()  # df processed inplace
                predictions = super().run(df, action=action, **kwargs)
            else:
                # run the pipeline with generic input return predictions
                predictions = super().run(*args, action=action, **kwargs)
            return predictions

        # let the factory raise the exception so it can fill it with context
        self.factory.exception("RecipePipelinePlugin doesn't implement action: %s", action, plugin=self)



"""
Plugins that group other plugins into logical groups like
ETL (extract, transform, load) pipeline or a graph used to
process data and create a machine learning model.
"""

import pandas as pd
import analitico.pandas

from analitico import status, AnaliticoException
from analitico.pandas import pd_to_dict
from analitico.utilities import time_ms
from analitico.schema import pandas_to_analitico_type, generate_schema
from analitico.constants import ACTION_PREDICT

from .interfaces import IGroupPlugin, plugin

##
## PipelinePlugin
##

DATAFRAME_SAMPLES = 10


@plugin
class PipelinePlugin(IGroupPlugin):
    """ 
    A plugin that creates a linear workflow by chaining together other plugins.
    Plugins that are chained in a pipeline need to take a single input and have
    a single output of the same kind so they same object can be processed from 
    the first, to the next and down to the last, then returned to caller as if
    the process was just one logical operation. PipelinePlugin can be used to 
    for example to construct ETL (extract, transform, load) workflows.
    """

    class Meta(IGroupPlugin.Meta):
        name = "analitico.plugin.PipelinePlugin"

    def get_metadata(self, *args):
        """ Transform list of arguments into a dictionary describing them (used to log status, etc) """
        output = []
        if args and len(args) > 0:
            for i, arg in enumerate(args):
                meta = {}
                meta["type"] = str(type(arg))
                if isinstance(arg, pd.DataFrame):
                    df = arg
                    meta["rows"] = len(df)
                    meta["schema"] = generate_schema(df)
                    samples = analitico.pandas.pd_sample(df, DATAFRAME_SAMPLES)
                    meta["samples"] = pd_to_dict(samples)

                    # debugging help
                    self.factory.debug("output[%d]: pd.DataFrame", i)
                    self.factory.debug("  rows: %d", len(df))
                    self.factory.debug("  columns: %d", len(df.columns))
                    for j, column in enumerate(df.columns):
                        self.factory.debug(
                            "  %3d %s (%s/%s)", j, column, df.dtypes[j], pandas_to_analitico_type(df.dtypes[j])
                        )
                else:
                    self.factory.debug("output[%d]: %s", i, str(type(arg)))
                output.append(meta)
        return output

    def run(self, *args, action=None, **kwargs):
        """ Process plugins in sequence, return combined result """
        try:
            pipeline_on = time_ms()

            # logging is expensive so we don't track everything in prediction mode
            predicting = action and ACTION_PREDICT in action
            if not predicting:
                self.factory.status(self, status.STATUS_RUNNING)

            for p, plugin in enumerate(self.plugins):
                plugin_on = time_ms()
                if not predicting:
                    self.factory.status(plugin, status.STATUS_RUNNING)

                # a plugin can have one or more input parameters and one or more
                # output parameters. results from a call to the next in the chain
                # are passed as tuples. when we finally return, if we have a single
                # result we unpackit, otherwise we return as tuple. this allows
                # a pipeline of plugins to chain plugins with a variable number of
                # parameters. each plugin is responsible for validating the type of
                # its input positional parameters and named parameters.
                try:
                    args = plugin.run(*args, action=action, **kwargs)
                    if not isinstance(args, tuple):
                        args = (args,)
                except Exception as e:
                    self.factory.status(plugin, status.STATUS_FAILED, exception=e)
                    raise

                # log outputs of plugin
                # TODO skip when predicting
                if not predicting:
                    output = self.get_metadata(*args)
                    self.factory.status(plugin, status.STATUS_COMPLETED, elapsed_ms=time_ms(plugin_on), output=output)

            if not predicting:
                # log outputs of pipeline
                self.factory.status(self, status.STATUS_COMPLETED, elapsed_ms=time_ms(pipeline_on), output=output)
            return args if len(args) > 1 else args[0]

        except Exception as e:
            self.factory.status(self, status.STATUS_FAILED)
            self.factory.exception(self.Meta.name + " failed while processing", item=self, exception=e)



import pandas as pd

import analitico.mixin

##
## DEPRECATED. VERSION USED BY SDK IS IN MODELS/DATASET
##


class Dataset(analitico.mixin.AttributeMixin):
    """ A dataset can retrieve data from a source and process it through a pipeline to generate a dataframe """

    plugin = None

    def __init__(self, factory=None, **kwargs):
        super().__init__(**kwargs)
        if "plugin" in kwargs:
            if not factory:
                raise Exception("Dataset should be initialized with a factory so it can create plugins")
            self.plugin = kwargs["plugin"]
            if isinstance(self.plugin, dict):
                self.plugin = factory.get_plugin(**self.plugin)

    def get_dataframe(self, **kwargs):
        """ Creates a pandas dataframe from the plugin of this dataset (usually a source or pipeline) """
        if self.plugin:
            df = self.plugin.run(**kwargs)
            assert isinstance(df, pd.DataFrame)
            return df
        return None



import unittest
import os.path
import pandas as pd
import random
import string
import io
import json

from analitico.factory import Factory
from analitico.schema import generate_schema

from .test_mixin import TestMixin

# pylint: disable=no-member

TITANIC_PUBLIC_URL = "https://storage.googleapis.com/public.analitico.ai/data/titanic/train.csv"


class FactoryTests(unittest.TestCase, TestMixin):
    """ Unit testing of Factory functionality: caching, creating items, plugins, etc """

    factory = Factory()

    def random_long_name(self):
        return "".join(random.choice(string.ascii_letters + string.digits) for _ in range(512))

    def test_factory_get_cache_filename_repeatable(self):
        unique_id = self.random_long_name()
        file1_0 = self.factory.get_cache_filename(unique_id)
        for _ in range(0, 20):
            file1_1 = self.factory.get_cache_filename(unique_id)
            self.assertEqual(file1_0, file1_1)

    def test_factory_get_cache_filename_unique(self):
        file1 = self.factory.get_cache_filename(self.random_long_name())
        for _ in range(2, 20):
            file2 = self.factory.get_cache_filename(self.random_long_name())
            self.assertNotEqual(file1, file2)

    def test_factory_get_cache_filename_shared(self):
        unique_id = self.random_long_name()
        factory1 = Factory()
        file1_1 = factory1.get_cache_filename(unique_id)
        factory2 = Factory()
        file1_2 = factory2.get_cache_filename(unique_id)
        self.assertEqual(file1_1, file1_2)

    def test_factory_get_cache_filename_in_cache(self):
        unique_id = self.random_long_name()
        file1 = self.factory.get_cache_filename(unique_id)
        self.assertTrue(file1.startswith(self.factory.get_cache_directory()))
        self.assertEqual(os.path.dirname(file1), self.factory.get_cache_directory())

    def test_factory_get_url_stream(self):
        stream = self.factory.get_url_stream(TITANIC_PUBLIC_URL)
        df = pd.read_csv(stream)

        self.assertEqual(len(df), 891)
        self.assertEqual(df.columns[1], "Survived")
        self.assertEqual(df.loc[0, "Name"], "Braund, Mr. Owen òèéàù Harris")

    def test_factory_get_gzip_url_stream_(self):
        # Server sends API responses as gzip streams
        stream = self.factory.get_url_stream("https://analitico.ai/api/runtime")
        data = json.load(stream)["data"]
        self.assertTrue("hardware" in data)
        self.assertTrue("platform" in data)
        self.assertTrue("python" in data)



import json
import os
import os.path

import analitico.dataset
import analitico.utilities
import analitico.plugin

from analitico.dataset import Dataset

ASSETS_PATH = os.path.dirname(os.path.realpath(__file__)) + "/assets"


class TestMixin:
    """ Basic unit testing functionality for analitico's tests """

    # Create default factory
    factory = analitico.authorize()

    def get_asset_path(self, path):
        """ Returns absolute path of file in test /assets directory """
        return os.path.join(ASSETS_PATH, path)

    def read_json_asset(self, path):
        with open(self.get_asset_path(path), "r") as f:
            text = f.read()
            text = text.replace("{assets}", ASSETS_PATH)
            return json.loads(text)

    def read_dataset_asset(self, path):
        json = self.read_json_asset(path)
        return Dataset(factory=self.factory, **json)

    def read_dataframe_asset(self, path):
        ds = self.read_dataset_asset(path)
        return ds.get_dataframe()

    def get_csv_plugin(self, **kwargs):
        name = analitico.plugin.CSV_DATAFRAME_SOURCE_PLUGIN
        return self.factory.get_plugin(name, **kwargs)



"""
Plugins that import dataframes from different sources
"""

import pandas
from analitico.utilities import get_dict_dot
from analitico.schema import analitico_to_pandas_type, apply_schema, NA_VALUES
from .interfaces import IDataframeSourcePlugin, PluginError, plugin

##
## CsvDataframeSourcePlugin
##


@plugin
class CsvDataframeSourcePlugin(IDataframeSourcePlugin):
    """ A plugin that returns a pandas dataframe from a csv file """

    class Meta(IDataframeSourcePlugin.Meta):
        name = "analitico.plugin.CsvDataframeSourcePlugin"

    def run(self, *args, **kwargs):
        """ Creates a pandas dataframe from the csv source """
        try:
            url = self.get_attribute("source.url")
            if not url:
                raise PluginError("URL of csv file cannot be empty.", plugin=self)

            # source schema is part of the source definition?
            schema = self.get_attribute("source.schema")

            # no schema was provided but the url is that of an analitico dataset in the cloud
            if not schema and url.startswith("analitico://") and url.endswith("/data/csv"):
                info_url = url.replace("/data/csv", "/data/info")
                info = self.factory.get_url_json(info_url)
                schema = get_dict_dot(info, "data.schema")

            # array of types for each column in the source
            columns = schema.get("columns") if schema else None

            dtype = None
            if columns:
                dtype = {}
                for column in columns:
                    if "type" in column:  # type is optionally defined
                        if column["type"] == "datetime":
                            dtype[column["name"]] = "object"
                        elif column["type"] == "timespan":
                            dtype[column["name"]] = "object"
                        else:
                            dtype[column["name"]] = analitico_to_pandas_type(column["type"])

            stream = self.factory.get_url_stream(url, binary=False)
            df = pandas.read_csv(stream, dtype=dtype, encoding="utf-8", na_values=NA_VALUES)

            tail = self.get_attribute("tail", 0)
            if tail > 0:
                rows_before = len(df)
                df = df.tail(tail)
                self.info("tail: %d, rows before: %d, rows after: %d", tail, rows_before, len(df))

            if schema:
                # reorder, filter, apply types, rename columns as requested in schema
                df = apply_schema(df, schema)

            return df
        except Exception as exc:
            self.exception("Error while processing: %s", url, exception=exc)
